package ssw.mj.test;

import ssw.mj.codegen.Code;

/**
 * Test cases for the <code>Parser</code> class.
 */
public class ParserTest extends CompilerTestCase {
	public void testShortestProgram () {
		init("program Test { void main() { } }");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void main(0) at 0");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 0);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);

		parseAndVerify();
	}
	
	public void testDefinitions() {
		init("program Test" + LF +
		     "  class myClass {" + LF +
		     "    int i;" + LF +
		     "    char c;" + LF +
		     "    int[] ia;" + LF +
		     "    myClass o;" + LF +
		     "    myClass[] oa;" + LF +
		     "  }" + LF +
		     "  final int fi = 20;" + LF +
		     "  final char fc = 'x';" + LF +
		     "  int gi;" + LF +
		     "  char gc;" + LF +
		     "  int[] gia;" + LF +
		     "  myClass go;" + LF +
		     "  myClass[] goa;" + LF +
		     "{" + LF +
		     "  myClass[] method() " + LF +
		     "    int i;" + LF +
		     "    char c;" + LF +
		     "    char[] ca;" + LF +
		     "    myClass o;" + LF +
		     "    myClass[] oa;" + LF +
		     "  {" + LF +
		     "    return null;" + LF +
		     "  }" + LF +
		     "  void main() { }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Type myClass: class(5)");
		expectSymTab("    Field 0: int i");
		expectSymTab("    Field 1: char c");
		expectSymTab("    Field 2: int[] ia");
		expectSymTab("    Field 3: class(5) o");
		expectSymTab("    Field 4: class(5)[] oa");
		expectSymTab("  Constant: int fi = 20");
		expectSymTab("  Constant: char fc = 'x'");
		expectSymTab("  Global Variable 0: int gi");
		expectSymTab("  Global Variable 1: char gc");
		expectSymTab("  Global Variable 2: int[] gia");
		expectSymTab("  Global Variable 3: class(5) go");
		expectSymTab("  Global Variable 4: class(5)[] goa");
		expectSymTab("  Method: class(5)[] method(0) at 0");
		expectSymTab("    Local Variable 0: int i");
		expectSymTab("    Local Variable 1: char c");
		expectSymTab("    Local Variable 2: char[] ca");
		expectSymTab("    Local Variable 3: class(5) o");
		expectSymTab("    Local Variable 4: class(5)[] oa");
		expectSymTab("  Method: void main(0) at " + (EX >= 6 ? "8" : "0"));
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 5);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectCode(EX >= 6, Code.trap,      (byte) 1);
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 0);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(5);
		expectMainPC(8);
		
		parseAndVerify();
	}
	
	public void testTypes() {
		init("program Test" + LF +
		     "  class C { int i1; }" + LF +
		     "{" + LF +
		     "  void main()" + LF +
		     "    int i1;" + LF +
		     "    int i2;" + LF +
		     "    C o1;" + LF +
		     "    C o2;" + LF +
		     "    C[] oa1;" + LF +
		     "    C[] oa2;" + LF +
		     "  {" + LF +
		     "    i1 = 0;" + LF +
		     "    i1 = i2;" + LF +
		     "    o1 = null;" + LF +
		     "    o1 = o2;" + LF +
		     "    oa1 = null;" + LF +
		     "    oa1 = oa2;" + LF +
		     "    oa1[i1] = o1;" + LF +
		     "    oa1[i1].i1 = i2;" + LF +
		     "    if (0 > i1) { }" + LF +
		     "    if (null != o1) { }" + LF +
		     "    if (null == oa1) { }" + LF +
		     "  }" + LF +
		     "}");
		
		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Type C: class(1)");
		expectSymTab("    Field 0: int i1");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: class(1) o1");
		expectSymTab("    Local Variable 3: class(1) o2");
		expectSymTab("    Local Variable 4: class(1)[] oa1");
		expectSymTab("    Local Variable 5: class(1)[] oa2");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 6);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.store_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.store_0);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.store_2);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 5, Code.store_2);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.store,     (byte) 4);
		expectCode(EX >= 5, Code.load,      (byte) 5);
		expectCode(EX >= 5, Code.store,     (byte) 4);
		expectCode(EX >= 5, Code.load,      (byte) 4);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.astore);
		expectCode(EX >= 5, Code.load,      (byte) 4);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.putfield,  (short) 0);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 6, Code.jle,       (short) 3);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 6, Code.jeq,       (short) 3);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.load,      (byte) 4);
		expectCode(EX >= 6, Code.jne,       (short) 3);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}
	
	public void testExprLocal() {
		init("program Test" + LF +
		     "  class C { int i; }" + LF +
		     "{" + LF +
		     "  void main()" + LF +
		     "    int i1, i2, i3, i4, i5;" + LF +
		     "  {" + LF +
		     "    i1 = i2;" + LF +
		     "    i1 += i2 + 3 * i3 - i4 % i5;" + LF +
		     "    i1++;" + LF +
		     "  }" + LF +
		     "}");
		
		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Type C: class(1)");
		expectSymTab("    Field 0: int i");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: int i3");
		expectSymTab("    Local Variable 3: int i4");
		expectSymTab("    Local Variable 4: int i5");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 5);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.store_0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.const_3);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.mul);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 5, Code.load,      (byte) 4);
		expectCode(EX >= 5, Code.rem);
		expectCode(EX >= 5, Code.sub);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.store_0);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) 1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testExprGlobal() {
		init("program Test" + LF +
		     "  class C { int i; }" + LF +
		     "  int i1, i2, i3, i4, i5;" + LF +
		     "{" + LF +
		     "  void main()" + LF +
		     "  {" + LF +
		     "    i1 = i2;" + LF +
		     "    i1 -= i2 + 3 * i3 - i4 % i5;" + LF +
		     "    i1--;" + LF +
		     "  }" + LF +
		     "}");
		
		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Type C: class(1)");
		expectSymTab("    Field 0: int i");
		expectSymTab("  Global Variable 0: int i1");
		expectSymTab("  Global Variable 1: int i2");
		expectSymTab("  Global Variable 2: int i3");
		expectSymTab("  Global Variable 3: int i4");
		expectSymTab("  Global Variable 4: int i5");
		expectSymTab("  Method: void main(0) at 0");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 0);
		expectCode(EX >= 5, Code.getstatic, (short) 1);
		expectCode(EX >= 5, Code.putstatic, (short) 0);
		expectCode(EX >= 5, Code.getstatic, (short) 0);
		expectCode(EX >= 5, Code.getstatic, (short) 1);
		expectCode(EX >= 5, Code.const_3);
		expectCode(EX >= 5, Code.getstatic, (short) 2);
		expectCode(EX >= 5, Code.mul);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.getstatic, (short) 3);
		expectCode(EX >= 5, Code.getstatic, (short) 4);
		expectCode(EX >= 5, Code.rem);
		expectCode(EX >= 5, Code.sub);
		expectCode(EX >= 5, Code.sub);
		expectCode(EX >= 5, Code.putstatic, (short) 0);
		expectCode(EX >= 5, Code.getstatic, (short) 0);
		expectCode(EX >= 5, Code.const_m1);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.putstatic, (short) 0);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(5);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testExprField() {
		init("program Test" + LF +
		     "  class C { int i; }" + LF +
		     "{" + LF +
		     "  void main()" + LF +
		     "    C i1, i2, i3, i4, i5;" + LF +
		     "  {" + LF +
		     "    i1.i = i2.i;" + LF +
		     "    i1.i *= i2.i + 3 * i3.i - i4.i % i5.i;" + LF +
		     "    i1.i++;" + LF +
		     "  }" + LF +
		     "}");
		
		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Type C: class(1)");
		expectSymTab("    Field 0: int i");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: class(1) i1");
		expectSymTab("    Local Variable 1: class(1) i2");
		expectSymTab("    Local Variable 2: class(1) i3");
		expectSymTab("    Local Variable 3: class(1) i4");
		expectSymTab("    Local Variable 4: class(1) i5");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 5);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.putfield,  (short) 0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.dup);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.const_3);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.mul);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.load,      (byte) 4);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.rem);
		expectCode(EX >= 5, Code.sub);
		expectCode(EX >= 5, Code.mul);
		expectCode(EX >= 5, Code.putfield,  (short) 0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.dup);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.const_1);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.putfield,  (short) 0);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testExprArray() {
		init("program Test" + LF +
		     "  class C { int i; }" + LF +
		     "{" + LF +
		     "  void main()" + LF +
		     "    C[] i1, i2, i3, i4, i5;" + LF +
		     "  {" + LF +
		     "    i1[1].i = i2[2].i;" + LF +
		     "    i1[1].i %= i2[2].i + 3 * i3[3].i - i4[4].i % i5[5].i;" + LF +
		     "    i1[1].i--;" + LF +
		     "  }" + LF +
		     "}");
		
		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Type C: class(1)");
		expectSymTab("    Field 0: int i");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: class(1)[] i1");
		expectSymTab("    Local Variable 1: class(1)[] i2");
		expectSymTab("    Local Variable 2: class(1)[] i3");
		expectSymTab("    Local Variable 3: class(1)[] i4");
		expectSymTab("    Local Variable 4: class(1)[] i5");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 5);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.const_1);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.const_2);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.putfield,  (short) 0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.const_1);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.dup);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.const_2);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.const_3);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.const_3);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.mul);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 5, Code.const_4);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.load,      (byte) 4);
		expectCode(EX >= 5, Code.const_5);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.rem);
		expectCode(EX >= 5, Code.sub);
		expectCode(EX >= 5, Code.rem);
		expectCode(EX >= 5, Code.putfield,  (short) 0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.const_1);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.dup);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.const_m1);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.putfield,  (short) 0);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}
	
	public void testConst() {
		init("program Test" + LF +
		     "  final int fi = 20;" + LF +
		     "  final char fc = 'a';" + LF +
		     "{" + LF +
		     "  void main()" + LF +
		     "    int i;" + LF +
		     "    char c;" + LF +
		     "  {" + LF +
		     "    i = 4;" + LF +
		     "    i = -1;" + LF +
		     "    i = 10;" + LF +
		     "    i = fi;" + LF +
		     "    i = -fi;" + LF +
		     "    i = fi + 5;" + LF +
		     "    i = 5 * (-fi);" + LF +
		     "    c = fc;" + LF +
		     "    c = 'a';" + LF +
		     "  }" + LF +
		     "}");
		
		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Constant: int fi = 20");
		expectSymTab("  Constant: char fc = 'a'");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i");
		expectSymTab("    Local Variable 1: char c");

		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 2);
		expectCode(EX >= 6, Code.const_4);
		expectCode(EX >= 6, Code.store_0);
		expectCode(EX >= 6, Code.const_m1);
		expectCode(EX >= 6, Code.store_0);
		expectCode(EX >= 6, Code.const_,    (int) 10);
		expectCode(EX >= 6, Code.store_0);
		expectCode(EX >= 6, Code.const_,    (int) 20);
		expectCode(EX >= 6, Code.store_0);
		expectCode(EX >= 6, Code.const_,    (int) -20);
		expectCode(EX >= 6, Code.store_0);
		expectCode(EX >= 6, Code.const_,    (int) 20);
		expectCode(EX >= 6, Code.const_5);
		expectCode(EX >= 6, Code.add);
		expectCode(EX >= 6, Code.store_0);
		expectCode(EX >= 6, Code.const_5);
		expectCode(EX >= 6, Code.const_,    (int) -20);
		expectCode(EX >= 6, Code.mul);
		expectCode(EX >= 6, Code.store_0);
		expectCode(EX >= 6, Code.const_,    (int) 97);
		expectCode(EX >= 6, Code.store_1);
		expectCode(EX >= 6, Code.const_,    (int) 97);
		expectCode(EX >= 6, Code.store_1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);

		parseAndVerify();
	}
	
	public void testNew() {
		init("program Test" + LF +
		     "  class C { int i; }" + LF +
		     "{" + LF +
		     "  void main()" + LF +
		     "    int i1, i2;" + LF +
		     "    C obj;" + LF +
		     "    char[] ca;" + LF +
		     "    C[] oa;" + LF +
		     "  {" + LF +
		     "    obj = new C;" + LF +
		     "    ca = new char[5];" + LF +
		     "    oa = new C[i1 * obj.i - oa[obj.i - 5].i];" + LF +
		     "  }" + LF +
		     "}");
		
		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Type C: class(1)");
		expectSymTab("    Field 0: int i");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: class(1) obj");
		expectSymTab("    Local Variable 3: char[] ca");
		expectSymTab("    Local Variable 4: class(1)[] oa");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 5);
		expectCode(EX >= 5, Code.new_,      (short) 4);
		expectCode(EX >= 5, Code.store_2);
		expectCode(EX >= 5, Code.const_5);
		expectCode(EX >= 5, Code.newarray,  (byte) 0);
		expectCode(EX >= 5, Code.store_3);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.mul);
		expectCode(EX >= 5, Code.load,      (byte) 4);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.const_5);
		expectCode(EX >= 5, Code.sub);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.sub);
		expectCode(EX >= 5, Code.newarray,  (byte) 1);
		expectCode(EX >= 5, Code.store,     (byte) 4);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testIf() {
		init("program Test {" + LF +
		     "  void main() int i1, i2, i3, i4; {" + LF +
		     "    if (i1 > i2) i1++; else i1--;" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: int i3");
		expectSymTab("    Local Variable 3: int i4");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 4);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 6, Code.jle,       (short) 9);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) 1);
		expectCode(EX >= 6, Code.jmp,       (short) 6);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) -1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testIfAnd() {
		init("program Test {" + LF +
		     "  void main() int i1, i2, i3, i4; {" + LF +
		     "    if (i1 > i2 && i3 < i4) i1++; else i1--;" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: int i3");
		expectSymTab("    Local Variable 3: int i4");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 4);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 6, Code.jle,       (short) 14);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 6, Code.jge,       (short) 9);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) 1);
		expectCode(EX >= 6, Code.jmp,       (short) 6);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) -1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testIfOr() {
		init("program Test {" + LF +
		     "  void main() int i1, i2, i3, i4; {" + LF +
		     "    if (i1 > i2 || i3 < i4) i1++; else i1--;" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: int i3");
		expectSymTab("    Local Variable 3: int i4");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 4);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 6, Code.jgt,       (short) 8);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 6, Code.jge,       (short) 9);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) 1);
		expectCode(EX >= 6, Code.jmp,       (short) 6);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) -1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testIfAndOr() {
		init("program Test {" + LF +
		     "  void main() int i1, i2, i3, i4; {" + LF +
		     "    if (i1 > i2 && i3 < i4 || i1 == i2 && i3 != i4) i1++; else i1--;" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: int i3");
		expectSymTab("    Local Variable 3: int i4");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 4);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 6, Code.jle,       (short) 8);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 6, Code.jlt,       (short) 13);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 6, Code.jne,       (short) 14);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 6, Code.jeq,       (short) 9);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) 1);
		expectCode(EX >= 6, Code.jmp,       (short) 6);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) -1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}
	
	public void testIfNested() {
		init("program Test {" + LF +
		     "  void main() int i1, i2, i3, i4; {" + LF +
		     "    if (i1 > i2) {" + LF +
		     "      if (i3 < i4) i1++;" + LF +
		     "    } else {" + LF +
		     "      if (i3 > i4) i1--;" + LF +
		     "    }" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: int i3");
		expectSymTab("    Local Variable 3: int i4");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 4);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 6, Code.jle,       (short) 14);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 6, Code.jge,       (short) 6);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) 1);
		expectCode(EX >= 6, Code.jmp,       (short) 11);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.load_3);
		expectCode(EX >= 6, Code.jle,       (short) 6);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) -1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testWhile() {
		init("program Test {" + LF +
		     "  void main() int i1, i2, i3, i4; {" + LF +
		     "    while (i1 > i2) {" + LF +
		     "      i1++;" + LF +
		     "    }" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: int i3");
		expectSymTab("    Local Variable 3: int i4");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 4);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 6, Code.jle,       (short) 9);
		expectCode(EX >= 5, Code.inc,       (byte) 0, (byte) 1);
		expectCode(EX >= 6, Code.jmp,       (short) -8);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testWhileNested() {
		init("program Test {" + LF +
		     "  void main() int i1, i2, i3, i4; {" + LF +
		     "    while (i1 > i2) {" + LF +
		     "      i1++;" + LF +
		     "      while (i3 < i4) {" + LF +
		     "        if (i1 == i2) break;" + LF +
		     "      }" + LF +
		     "      if (i3 != i4) break;" + LF +
		     "    }" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i1");
		expectSymTab("    Local Variable 1: int i2");
		expectSymTab("    Local Variable 2: int i3");
		expectSymTab("    Local Variable 3: int i4");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 4);
		expectCode(EX >= 6, Code.load_0);
		expectCode(EX >= 6, Code.load_1);
		expectCode(EX >= 6, Code.jle,       (short) 33);
		expectCode(EX >= 6, Code.inc,       (byte) 0, (byte) 1);
		expectCode(EX >= 6, Code.load_2);
		expectCode(EX >= 6, Code.load_3);
		expectCode(EX >= 6, Code.jge,       (short) 14);
		expectCode(EX >= 6, Code.load_0);
		expectCode(EX >= 6, Code.load_1);
		expectCode(EX >= 6, Code.jne,       (short) 6);
		expectCode(EX >= 6, Code.jmp,       (short) 6);
		expectCode(EX >= 6, Code.jmp,       (short) -13);
		expectCode(EX >= 6, Code.load_2);
		expectCode(EX >= 6, Code.load_3);
		expectCode(EX >= 6, Code.jeq,       (short) 6);
		expectCode(EX >= 6, Code.jmp,       (short) 6);
		expectCode(EX >= 6, Code.jmp,       (short) -32);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
		
		parseAndVerify();
	}

	public void testPredefMeth() {
		init("program Test {" + LF +
		     "  void main() int i; char c; int[] ia; {" + LF +
		     "    i = ord(c);" + LF +
		     "    c = chr(i);" + LF +
		     "    i = len(ia);" + LF +
		     "    print(i);" + LF +
		     "    print(c, 4);" + LF +
		     "    read(i);" + LF +
		     "    read(c);" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int i");
		expectSymTab("    Local Variable 1: char c");
		expectSymTab("    Local Variable 2: int[] ia");
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 3);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.store_0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.store_1);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 6, Code.arraylength);
		expectCode(EX >= 5, Code.store_0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.print);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.const_4);
		expectCode(EX >= 5, Code.bprint);
		expectCode(EX >= 5, Code.read);
		expectCode(EX >= 5, Code.store_0);
		expectCode(EX >= 5, Code.bread);
		expectCode(EX >= 5, Code.store_1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(0);
	
		parseAndVerify();
	}
	
	public void testCallVoid() {
		init("program Test {" + LF +
		     "  void method() {" + LF +
		     "    return;" + LF +
		     "  }" + LF +
		     "  void main() {" + LF +
		     "    method();" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: void method(0) at 0");
		expectSymTab("  Method: void main(0) at " + (EX >= 6 ? "7" : "0"));
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 0);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 0);
		expectCode(EX >= 6, Code.call,      (short) -10);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(7);

		parseAndVerify();
	}

	public void testCallInt() {
		init("program Test {" + LF +
		     "  int method() {" + LF +
		     "    return 1;" + LF +
		     "  }" + LF +
		     "  void main() {" + LF +
		     "    print(method());" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: int method(0) at 0");
		expectSymTab("  Method: void main(0) at " + (EX >= 6 ? "8" : "0"));
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 0);
		expectCode(EX >= 5, Code.const_1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectCode(EX >= 6, Code.trap,      (byte) 1);
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 0);
		expectCode(EX >= 6, Code.call,      (short) -11);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.print);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(8);

		parseAndVerify();
	}

	public void testCallParam() {
		init("program Test {" + LF +
		     "  int method(int i, char c, int[] ia) {" + LF +
		     "    return i + ia[ord(c)];" + LF +
		     "  }" + LF +
		     "  void main() int i; char c; {" + LF +
		     "    print(method(i, c, new int[i]));" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: int method(3) at 0");
		expectSymTab("    Local Variable 0: int i");
		expectSymTab("    Local Variable 1: char c");
		expectSymTab("    Local Variable 2: int[] ia");
		expectSymTab("  Method: void main(0) at " + (EX >= 6 ? "12" : "0"));
		expectSymTab("    Local Variable 0: int i");
		expectSymTab("    Local Variable 1: char c");
		
		expectCode(EX >= 6, Code.enter,     (byte) 3, (byte) 3);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_2);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectCode(EX >= 6, Code.trap,      (byte) 1);
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 2);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.newarray,  (byte) 1);
		expectCode(EX >= 6, Code.call,      (short) -20);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.print);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(12);

		parseAndVerify();
	}

	public void testCallFuncAsProc() {
		init("program Test {" + LF +
		     "  int method() {" + LF +
		     "    return 1;" + LF +
		     "  }" + LF +
		     "  void main() {" + LF +
		     "    method();" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: int method(0) at 0");
		expectSymTab("  Method: void main(0) at " + (EX >= 6 ? "8" : "0"));
		
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 0);
		expectCode(EX >= 5, Code.const_1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectCode(EX >= 6, Code.trap,      (byte) 1);
		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 0);
		expectCode(EX >= 6, Code.call,      (short) -11);
		expectCode(EX >= 6, Code.pop);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(8);

		parseAndVerify();
	}

	public void testMultReturn() {
		init("program Test {" + LF +
		     "  int method(int p) {" + LF +
		     "    if (p > 0) return 1;" + LF +
		     "    if (p < 0) return -1;" + LF +
		     "    return 0;" + LF +
		     "  }" + LF +
		     "  void main() {" + LF +
		     "    print(method(5));" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program Test:");
		expectSymTab("  Method: int method(1) at 0");
		expectSymTab("    Local Variable 0: int p");
		expectSymTab("  Method: void main(0) at " + (EX >= 6 ? "24" : "0"));
		
		expectCode(EX >= 6, Code.enter,     (byte) 1, (byte) 1);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 6, Code.jle,       (short) 6);
		expectCode(EX >= 5, Code.const_1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 6, Code.jge,      (short) 6);
		expectCode(EX >= 5, Code.const_m1);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectCode(EX >= 6, Code.trap,     (byte) 1);
		expectCode(EX >= 6, Code.enter,    (byte) 0, (byte) 0);
		expectCode(EX >= 5, Code.const_5);
		expectCode(EX >= 6, Code.call,     (short) -28);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.print);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(0);
		expectMainPC(24);

		parseAndVerify();
	}

	
	public void testScriptExample() {
		init("program P" + LF +
		     "  final int size = 10;" + LF +
		     "" + LF +
		     "  class Table {" + LF +
		     "    int[] pos;" + LF +
		     "    int[] neg;" + LF +
		     "  }" + LF +
		     "" + LF +
		     "  Table val;" + LF +
		     "" + LF +
		     "{" + LF +
		     "  void main()" + LF +
		     "    int x, i;" + LF +
		     "  { /*---------- Initialize val */" + LF +
		     "    val = new Table;" + LF +
		     "    val.pos = new int[size];" + LF +
		     "    val.neg = new int[size];" + LF +
		     "    i = 0;" + LF +
		     "    while (i < size) {" + LF +
		     "      val.pos[i] = 0; val.neg[i] = 0;" + LF +
		     "      i++;" + LF +
		     "    }" + LF +
		     "    /*---------- Read values */" + LF +
		     "    read(x);" + LF +
		     "    while (x != 0) {" + LF +
		     "      if (0 <= x && x < size) {" + LF +
		     "        val.pos[x]++;" + LF +
		     "      } else if (-size < x && x < 0) {" + LF +
		     "        val.neg[-x]++;" + LF +
		     "      }" + LF +
		     "      read(x);" + LF +
		     "    }" + LF +
		     "  }" + LF +
		     "}");

		expectSymTabUniverse();
		expectSymTab("Program P:");
		expectSymTab("  Constant: int size = 10");
		expectSymTab("  Type Table: class(2)");
		expectSymTab("    Field 0: int[] pos");
		expectSymTab("    Field 1: int[] neg");
		expectSymTab("  Global Variable 0: class(2) val");
		expectSymTab("  Method: void main(0) at 0");
		expectSymTab("    Local Variable 0: int x");
		expectSymTab("    Local Variable 1: int i");

		expectCode(EX >= 6, Code.enter,     (byte) 0, (byte) 2);
		expectCode(EX >= 5, Code.new_,      (short) 8);
		expectCode(EX >= 5, Code.putstatic, (short) 0);
		expectCode(EX >= 5, Code.getstatic, (short) 0);
		expectCode(EX >= 5, Code.const_,    (int) 10);
		expectCode(EX >= 5, Code.newarray,  (byte) 1);
		expectCode(EX >= 5, Code.putfield,  (short) 0);
		expectCode(EX >= 5, Code.getstatic, (short) 0);
		expectCode(EX >= 5, Code.const_,    (int) 10);
		expectCode(EX >= 5, Code.newarray,  (byte) 1);
		expectCode(EX >= 5, Code.putfield,  (short) 1);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.store_1);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.const_,    (int) 10);
		expectCode(EX >= 6, Code.jge,       (short) 27);
		expectCode(EX >= 5, Code.getstatic, (short) 0);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.astore);
		expectCode(EX >= 5, Code.getstatic, (short) 0);
		expectCode(EX >= 5, Code.getfield,  (short) 1);
		expectCode(EX >= 5, Code.load_1);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.astore);
		expectCode(EX >= 5, Code.inc,       (byte) 1, (byte) 1);
		expectCode(EX >= 6, Code.jmp,       (short) -30);
		expectCode(EX >= 5, Code.read);
		expectCode(EX >= 5, Code.store_0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 6, Code.jeq,       (short) 64);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 6, Code.jgt,       (short) 27);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.const_,    (int) 10);
		expectCode(EX >= 6, Code.jge,       (short) 18);
		expectCode(EX >= 5, Code.getstatic, (short) 0);
		expectCode(EX >= 5, Code.getfield,  (short) 0);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.dup2);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.const_1);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.astore);
		expectCode(EX >= 6, Code.jmp,       (short) 30);
		expectCode(EX >= 5, Code.const_,    (int) -10);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 6, Code.jge,       (short) 21);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.const_0);
		expectCode(EX >= 6, Code.jge,       (short) 16);
		expectCode(EX >= 5, Code.getstatic, (short) 0);
		expectCode(EX >= 5, Code.getfield,  (short) 1);
		expectCode(EX >= 5, Code.load_0);
		expectCode(EX >= 5, Code.neg);
		expectCode(EX >= 5, Code.dup2);
		expectCode(EX >= 5, Code.aload);
		expectCode(EX >= 5, Code.const_1);
		expectCode(EX >= 5, Code.add);
		expectCode(EX >= 5, Code.astore);
		expectCode(EX >= 5, Code.read);
		expectCode(EX >= 5, Code.store_0);
		expectCode(EX >= 6, Code.jmp,       (short) -63);
		expectCode(EX >= 6, Code.exit);
		expectCode(EX >= 6, Code.return_);
		expectDataSize(1);
		expectMainPC(0);

		parseAndVerify();
	}
}
